//	3{3ZT[MPU6050Cu
//	wt]xH쎺	http://jsdiy.webcrow.jp
//	2018/07

#include <avr/io.h>
#include <math.h>
#include "twilib.h"
#include "mpu6050lib.h"

//MPU6050̃X[uAhX
#define	MPU6050_SLAVE_ADDRESS	0x68

//WX^`
enum ERegAddr
{
	MPU6050_ACCEL_XOUT_H = 0x3B,
	MPU6050_ACCEL_XOUT_L,
	MPU6050_ACCEL_YOUT_H,
	MPU6050_ACCEL_YOUT_L,
	MPU6050_ACCEL_ZOUT_H,
	MPU6050_ACCEL_ZOUT_L,
	MPU6050_TEMP_OUT_H,
	MPU6050_TEMP_OUT_L,
	MPU6050_GYRO_XOUT_H,
	MPU6050_GYRO_XOUT_L,
	MPU6050_GYRO_YOUT_H,
	MPU6050_GYRO_YOUT_L,
	MPU6050_GYRO_ZOUT_H,
	MPU6050_GYRO_ZOUT_L,
	
	MPU6050_GYRO_CONFIG = 0x1B,
	MPU6050_ACCEL_CONFIG = 0x1C,
	
	//The reset value is 0x00 for all registers other than the registers below.
	MPU6050_PWR_MGMT_1 = 0x6B,	//reset value is 0x40
	MPU6050_WHO_AM_I = 0x75		//reset value is 0x68
};

#define MPU6050_DATA_START_ADDR	MPU6050_ACCEL_XOUT_H

//WX^l\
typedef	struct
{
	/*
	int16_t accelX;
	int16_t accelY;
	int16_t accelZ;
	int16_t temperature;
	int16_t gyroX;
	int16_t gyroY;
	int16_t gyroZ;
	*/
	TIAxis	accel;
	int16_t temperature;
	TIAxis	gyro;
}
TRegValue;

//vg^Cv錾
static	void	WriteRegValue(uint8_t regAddr, uint8_t data);
static	void	ReadRegValue(uint8_t startAddr, uint8_t* readBuf, uint8_t readLength);
static	void	SwapByte(int16_t* value);

static	float	gyroLsbSens, accelLsbSens;

//
uint8_t	MPU6050_Init(void)
{
	return MPU6050_Init2(MPU6050_GYRO_SCALE_DEFAULT, MPU6050_ACCEL_SCALE_DEFAULT);
}

uint8_t	MPU6050_Init2(EMpu6050GyroScale gscale, EMpu6050AccelScale ascale)
{
	uint8_t	data;
	
	//ʐMmF
	ReadRegValue(MPU6050_WHO_AM_I, &data, 1);
	if (data != 0x68) { return 0; }
	
	//Ԃݒ肷
	/*	foCXZbg	bit7:DEVICE_RESET = 0	//'1'ƑSWX^NA
		X[v[h		bit6:SLEEP = 0			//'0'ŉ
		ITvO	bit5:CYCLE = 0			//vOŏȃ^C~OŌĂяô
		i󂫁j			bit4:Ȃ
		xZT[	bit3:TEMP_DIS = 0		//'1'ƁuxZT[vIƂȂ
		NbN\[X		bit210:CLKSEL = 000		//'000'MPU-6050̓8MHz
	*/
	WriteRegValue(MPU6050_PWR_MGMT_1, 0x00);
	
	//𑜓xݒ肷iWCj
	if (gscale != MPU6050_GYRO_SCALE_DEFAULT)
	{
		ReadRegValue(MPU6050_GYRO_CONFIG, &data, 1);
		data = (data & 0xE7) | (gscale << 3);
		WriteRegValue(MPU6050_GYRO_CONFIG, data);
	}
	
	const	float	GyroLsbSensList[] = { 131.0, 65.5, 32.8, 16.4 };
	gyroLsbSens = GyroLsbSensList[gscale];
	
	//𑜓xݒ肷ixj
	if (ascale != MPU6050_ACCEL_SCALE_DEFAULT)
	{
		ReadRegValue(MPU6050_ACCEL_CONFIG, &data, 1);
		data = (data & 0xE7) | (ascale << 3);
		WriteRegValue(MPU6050_ACCEL_CONFIG, data);
	}
	
	const	float	AccelLsbSensList[] = { 16384.0, 8192.0, 4096.0, 2048.0 };
	accelLsbSens = AccelLsbSensList[ascale];
	
	return 1;
}

//ZT[l擾
void	MPU6050_GetValue(TMpu6050Value* retValue)
{
	TRegValue	reg;

	ReadRegValue(MPU6050_DATA_START_ADDR, (uint8_t*)&reg, sizeof(TRegValue));
	SwapByte(&reg.accel.x);
	SwapByte(&reg.accel.y);
	SwapByte(&reg.accel.z);
	SwapByte(&reg.temperature);
	SwapByte(&reg.gyro.x);
	SwapByte(&reg.gyro.y);
	SwapByte(&reg.gyro.z);

	/*	yMPU-60x0-Register-Mapz	4.17 Registers 59 to 64 - Accelerometer Measurements
		[Full Scale Range]	[LSB Sensitivity]
		}2g				16384 LSB/g			iftHg̃WX^l̏ꍇj
	*/
	//d͉xG̉{̉xĂ邩
	float faccelX = (float)reg.accel.x / accelLsbSens;
	float faccelY = (float)reg.accel.y / accelLsbSens;
	float faccelZ = (float)reg.accel.z / accelLsbSens;
	
	//ʂɑ΂ZT[̌X
	/*	yQlz
		http://www.analog.com/media/jp/technical-documentation/application-notes/AN-1057_jp.pdf
		Page6-7u3̌X΂̌vZv̐}
	*/
	retValue->angle.x =  180.0 / M_PI * atan2(faccelY, hypot(faccelX, faccelZ));	//X𒆐SƂ]pi}̃Ձj
	retValue->angle.y = -180.0 / M_PI * atan2(faccelX, hypot(faccelY, faccelZ));	//Y𒆐SƂ]pi}̃Ɓj	]͓W[̐ƍ킹邽
	retValue->angle.z =  180.0 / M_PI * atan2(hypot(faccelX, faccelY), faccelZ);	//ZƏd͕Ƃ̂Ȃp
	
	/*	yMPU-60x0-Register-Mapz	4.19 Registers 67 to 72 - Gyroscope Measurements
		[Full Scale Range]	[LSB Sensitivity]
		} 250 /s			131 LSB//s		iftHg̃WX^l̏ꍇj
		EpxBb250x܂ł̑ɑΉB1]ƒl131B
		E+-250 * 131 = +-32760  int16_tŕ\\Ȕ͈́B
	*/
	//bx̑ŉ]Ă邩(degree/sec)
	retValue->degps.x = (float)reg.gyro.x / gyroLsbSens;	//]̐̌̓W[̐}̒ʂ
	retValue->degps.y = (float)reg.gyro.y / gyroLsbSens;	//]̐̌̓W[̐}̒ʂ
	retValue->degps.z = (float)reg.gyro.z / gyroLsbSens;	//]͔̐̌v
}

//wAhXɃf[^
static	void	WriteRegValue(uint8_t regAddr, uint8_t data)
{
	TWI_Start();
	TWI_SlaveAddrW(MPU6050_SLAVE_ADDRESS);
	TWI_WriteByte(regAddr);
	TWI_WriteByte(data);
	TWI_Stop();
}

//wAhXf[^ǂ
static	void	ReadRegValue(uint8_t startAddr, uint8_t* readBuf, uint8_t readLength)
{
	TWI_Start();
	TWI_SlaveAddrW(MPU6050_SLAVE_ADDRESS);
	TWI_WriteByte(startAddr);
	TWI_Stop();
	
	TWI_Start();
	TWI_SlaveAddrR(MPU6050_SLAVE_ADDRESS);
	for (uint8_t i = 0; i < readLength; i++)
	{
		TWI_ReadByte((i == readLength - 1), readBuf + i);
	}
	TWI_Stop();
}

//MPU6050ǂݍ2bytel̓GfBAAVRƂ͋tȂ̂ŏ㉺byteւ
static	void	SwapByte(int16_t* value)
{
	*value = ((*value << 8) & 0xFF00) | ((*value >> 8) & 0x00FF);
}
